//
//  PSMLanListener.m
//  BombSquad
//
//  Created by John Pannell on 2/15/10.
//  Copyright 2010 Positive Spin Media. All rights reserved.
//

#import "PSMLanListener.h"
#import <sys/socket.h>
#import <netinet/in.h>
#import <unistd.h>

#pragma mark -
#pragma mark CFNetwork C Callbacks

static void lanListenerAcceptCallback (CFSocketRef theSocket, CFSocketCallBackType theType, CFDataRef theAddress, const void *data, void *info){
    
    // set up
    PSMLanListener *listener = (PSMLanListener *)info;
    id listenerDelegate = [listener delegate];
    
    // is this the right kind of notification?
    if(theType == kCFSocketAcceptCallBack){
        CFSocketNativeHandle nativeSocket = *(CFSocketNativeHandle *)data;
        uint8_t name[SOCK_MAXADDRLEN];
        socklen_t namelen = sizeof(name);
        NSData *peerAddress = nil;
        
        // identify the other end of the line
        if(getpeername(nativeSocket, (struct sockaddr *)name, &namelen) == 0){
            peerAddress = [NSData dataWithBytes:name length:namelen];
        }
        
        // get streams
        CFReadStreamRef readStream = NULL;
        CFWriteStreamRef writeStream = NULL;
        CFStreamCreatePairWithSocket(kCFAllocatorDefault, nativeSocket, &readStream, &writeStream);
        
        // worked? set them up and hand off...
        if(readStream && writeStream){
            CFReadStreamSetProperty(readStream, kCFStreamPropertyShouldCloseNativeSocket, kCFBooleanTrue);
            CFWriteStreamSetProperty(writeStream, kCFStreamPropertyShouldCloseNativeSocket, kCFBooleanTrue);
            if(listenerDelegate && [listenerDelegate respondsToSelector:@selector(acceptedConnectionForListener:inputStream:outputStream:address:)]){
                [listenerDelegate acceptedConnectionForListener:listener inputStream:(NSInputStream *)readStream outputStream:(NSOutputStream *)writeStream address:peerAddress];
            }
        } else { // no streams? error out...
            NSLog(@"listener did not get streams");
            close(nativeSocket);
            if(listenerDelegate && [listenerDelegate respondsToSelector:@selector(lanListener:encounteredError:)]){
                NSError *error = [[NSError alloc] initWithDomain:kPSMLanListenerErrorDomain code:kPSMLanListenerErrorStreamError userInfo:nil];
                [listenerDelegate lanListener:listener encounteredError:error];
                [error release];
            }
        }
        
        // clean up memory
        if(readStream) CFRelease(readStream);
        if(writeStream) CFRelease(writeStream);
    }
}

@implementation PSMLanListener
@synthesize delegate, port;

#pragma mark -
#pragma mark Creation/Destruction

- (id)init
{
    if(self = [super init]){
        socket = NULL;
        delegate = nil;
        port = -1;
    }
    return self;
}

- (void)dealloc
{
    [self stopListening];
    [super dealloc];
}

#pragma mark -
#pragma mark Listener Methods

- (BOOL)startListening:(NSError **)error
{
    // configure a socket for listening
    CFSocketContext socketCtxt = {0, self, NULL, NULL, NULL};
    socket = CFSocketCreate(kCFAllocatorDefault, PF_INET, SOCK_STREAM, IPPROTO_TCP, kCFSocketAcceptCallBack, (CFSocketCallBack)&lanListenerAcceptCallback, &socketCtxt);
    
    // did it work?  error out if not
    if(socket == NULL){
        if(error){
            *error = [[NSError alloc] initWithDomain:kPSMLanListenerErrorDomain code:kPSMLanListenerErrorNoSocketsAvailable userInfo:nil];
        }
        return NO;
    }
    
    // set up socket to reuse address
    int ret = 1;
    setsockopt(CFSocketGetNative(socket), SOL_SOCKET, SO_REUSEADDR, (void *)&ret, sizeof(ret));
    
    // bind to an address
    struct sockaddr_in addr4;
    memset(&addr4, 0, sizeof(addr4));
    addr4.sin_len = sizeof(addr4);
    addr4.sin_family = AF_INET;
    addr4.sin_port = 0; // let the kernel pick the port
    addr4.sin_addr.s_addr = htonl(INADDR_ANY);
    NSData *address4 = [NSData dataWithBytes:&addr4 length:sizeof(addr4)];
    if(kCFSocketSuccess != CFSocketSetAddress(socket, (CFDataRef)address4)){
        if(error){
            *error = [[NSError alloc] initWithDomain:kPSMLanListenerErrorDomain code:kPSMLanListenerErrorCouldntBindToAddress userInfo:nil];
        }
        if(socket) CFRelease(socket);
        return NO;
    }
    
    // record the port number
    NSData *addr = [(NSData *)CFSocketCopyAddress(socket) autorelease];
    memcpy(&addr4, [addr bytes], [addr length]);
    port = ntohs(addr4.sin_port);
    
    // register with the run loop
    CFRunLoopRef cfrl = CFRunLoopGetCurrent();
    CFRunLoopSourceRef source4 = CFSocketCreateRunLoopSource(kCFAllocatorDefault, socket, 0);
    CFRunLoopAddSource(cfrl, source4, kCFRunLoopCommonModes);
    CFRelease(source4);
    
    return ret;
}

- (void)stopListening
{
    if(socket){
        CFSocketInvalidate(socket);
        CFRelease(socket);
        socket = NULL;
    }
}

@end
