//
//  PSMLanPeer.m
//  BombSquad
//
//  Created by John Pannell on 2/15/10.
//  Copyright 2010 Positive Spin Media. All rights reserved.
//

#import "PSMLanPeer.h"
#import "PSMLanProtocols.h"
#import "PSMPacketCategories.h"

@interface PSMLanPeer()

@property(readwrite) PSMLanPeerConnectionState connectionState;
@property(readwrite, retain) NSString *displayName;
@property(readwrite, retain) NSString *peerID;

// private - internal communication
- (BOOL)internalSendData:(NSData *)data;
- (BOOL)sendMessage:(NSInteger)message withPayload:(NSData *)payload;
- (void)handleMessage:(NSInteger)message withPayload:(NSData *)payload;
- (void)acceptRemoteConnection;
- (void)denyRemoteConnection;
- (void)disconnectRemote;
- (void)requestRemoteACK;
- (void)sendACKResponse;
- (void)revertState;
- (void)killRequestTimer;

// connection requests
- (void)connectionRequestTimedOut:(NSTimer *)timer;

// actual transmission across the wire
- (void)sendQueuedData;
- (void)closeStreams;

@end

@implementation PSMLanPeer
@synthesize displayName, peerID, remoteService, inputStream, outputStream, errorCode;
@dynamic connectionState;

#pragma mark -
#pragma mark Creation/Destruction

/* 
 
 Creation based on a discovery on the network.  We'll resolve the netservice, fetch the peer ID from the TXT record, and note the availability of this peer.
 
 */
- (id)initWithNetService:(NSNetService *)ns session:(PSMLanSession *)object
{
    if(self = [super init]){
        remoteService = [ns retain];
        [remoteService setDelegate:self];
        
        inputStream = nil;
        outputStream = nil;
        
        packetQueue = [[NSMutableArray alloc] init];
        
        // delegate pattern
        session = object;
        
        peerID = [[ns name] copy];
        connectionState = PSMLanPeerStateUnavailable;
        errorCode = 0;
        displayName = nil;
        
        writeReady = NO;
        readReady = NO;
        acknowledged = YES; // assumed good
        
        // resolve service in order to obtain the display name - will hear back in delegate methods
        [ns resolveWithTimeout:5];
        
    }
    
    return self;
    
}

- (void)dealloc
{
    [self closeStreams];
    
    if(remoteService){
        [remoteService release];
    }
    
    [packetQueue release];
    [writeLeftover release];
    [readLeftover release];
    
    [peerID release];
    [displayName release];
    if(requestTimer){
        [requestTimer invalidate];
        [requestTimer release];
        requestTimer = nil;
    }
    
    [super dealloc];
}

#pragma mark -
#pragma mark Custom Accessors

- (void)setConnectionState:(PSMLanPeerConnectionState)newState
{
    connectionState = newState;
    [session peer:self didChangeState:newState];
}

- (PSMLanPeerConnectionState)connectionState
{
    return connectionState;
}

#pragma mark -
#pragma mark Connecting
#pragma mark (starting a connection)

// client-side: ask for the connection and then wait for the response...
- (void)initiateConnectionWithTimeout:(NSTimeInterval)interval
{
    // use discovered net service to connect
    [self setConnectionState:PSMLanPeerStateConnecting];
    if([remoteService getInputStream:&inputStream outputStream:&outputStream]){
        [inputStream retain]; // I have read memory management on this both ways - run loop is holding these
        [outputStream retain];// I'm going to hold them too (as if inited with them) so dealloc can be consistent
        
        // open them up
        [inputStream setDelegate:self];
        [outputStream setDelegate:self];
        [inputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        [outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        [inputStream open];
        [outputStream open];
        
        // I will now wait accept/deny from other end
        requestTimer = [NSTimer scheduledTimerWithTimeInterval:interval target:self selector:@selector(connectionRequestTimedOut:) userInfo:nil repeats:NO];
        [requestTimer retain];
    } else {
        // non si connecti
        NSError *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionConnectionFailedError userInfo:nil];
        [session connectionRequestWithPeerFailed:self withError:error];
        [self setConnectionState:PSMLanPeerStateAvailable];
        NSLog(@"did not get streams");
    }
}

- (void)connectionRequestTimedOut:(NSTimer *)timer
{
    if(connectionState == PSMLanPeerStateConnecting){
        // tell the other end
        uint16_t code[1];
        code[0] = (uint16_t)PSMLanSessionTimedOutError;
        NSData *sendData = [NSData dataWithBytes:code length:sizeof(uint16_t)];
        if(![self sendMessage:kPSMPeerMessageSetErrorCode withPayload:sendData])
            NSLog(@"Failed to send error code (timed out) to remote peer");
        
        // tell this end
        NSError *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionTimedOutError userInfo:nil];
        if(session)
            [session connectionRequestWithPeerFailed:self withError:error];
        
        [self closeStreams];
        [self revertState];
    }
}

- (void)cancelConnectionRequest
{
    if(connectionState == PSMLanPeerStateConnecting){
        // tell the other end
        uint16_t code[1];
        code[0] = (uint16_t)PSMLanSessionCancelledError;
        NSData *sendData = [NSData dataWithBytes:code length:sizeof(uint16_t)];
        if(![self sendMessage:kPSMPeerMessageSetErrorCode withPayload:sendData])
            NSLog(@"Failed to send error code (cancelled) to remote peer");
        
        // tell this end
        NSError *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionCancelledError userInfo:nil];
        [session connectionRequestWithPeerFailed:self withError:error];
        
        [self closeStreams];
        [self revertState];
    }
}

#pragma mark (receiving a connection request)
- (BOOL)acceptConnection:(NSError **)error
{
    if(errorCode == 0){
        [self acceptRemoteConnection];
        [self setConnectionState:PSMLanPeerStateConnected];
        return YES;
    } else {
        // FAILED - could have timed out or been cancelled on other side
        NSLog(@"accpt connection failure: %d", errorCode);
        if(error != NULL){
            *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:errorCode userInfo:nil];
        }
        errorCode = 0;
        [self closeStreams];
        [self setConnectionState:PSMLanPeerStateAvailable];
        return NO;
    }
}

- (void)denyConnection
{
    [self denyRemoteConnection];
    [self closeStreams];
    [self setConnectionState:PSMLanPeerStateAvailable];
}

#pragma mark (ending a connection)

- (void)disconnect
{
    [self disconnectRemote];
    [self closeStreams];
    [self setConnectionState:PSMLanPeerStateDisconnected];
    [self revertState];
}

- (void)kickOut // this is disconnecting the other end from everyone
{
    if(![self sendMessage:kPSMPeerMessageKickOutRemote withPayload:nil])
        NSLog(@"Failed to send data to kick out remote connection");
}

- (void)hide // shortcut for session to set a particular connection state
{
    [self setConnectionState:PSMLanPeerStateUnavailable];
}

#pragma mark -
#pragma mark Usage

- (BOOL)sendData:(NSData *)data error:(NSError **)error
{
    if(data == nil || [data length] == 0){
        *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionInvalidParameterError userInfo:nil];
        return NO;
    }
    
    if(connectionState != PSMLanPeerStateConnected){
        *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionNotConnectedError userInfo:nil];
        return NO;
    }
    
    // PUT A ONE BYTE HEADER ON PUBLICLY SENT DATA PACKETS TO IDENTIFY IT AS EXTERNAL, THEN SEND
    NSMutableData *sendData = [NSMutableData data];
    uint8_t signal[1];
    signal[0] = kPSMPeerMessageExternal;
    [sendData appendBytes:signal length:sizeof(uint8_t)];
    [sendData appendBytes:[data bytes] length:[data length]];
    
    [packetQueue addObject:sendData];
    if([outputStream hasSpaceAvailable])
        [self sendQueuedData];
    
    return YES;
}

#pragma mark -
#pragma mark Private

- (void)testRemotePresence
{
    acknowledged = NO;
    [self requestRemoteACK];
    [self performSelector:@selector(testResults) withObject:nil afterDelay:2.0];
}

- (void)testResults
{
    if(!acknowledged){
        [self setConnectionState:PSMLanPeerStateDisconnected];
        NSError *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionNotConnectedError userInfo:nil]; 
        [session peer:self encounteredReadWriteError:error];
    }
}

- (void)revertState
{
    if([[session availablePeers] containsObject:[self peerID]]){
        [self setConnectionState:PSMLanPeerStateAvailable];
    } else {
        [self setConnectionState:PSMLanPeerStateUnavailable];
    }
}

- (void)killRequestTimer
{
    if(requestTimer){
        [requestTimer invalidate];
        [requestTimer release];
        requestTimer = nil;
    }
}

- (void)sendQueuedData
{
    if(writeLeftover == nil && [packetQueue count] == 0)
        return; // nothing to send
    
    NSMutableData *dataToSend = [NSMutableData data];
    if(writeLeftover != nil){
        [dataToSend appendData:writeLeftover];
        [writeLeftover release];
        writeLeftover = nil;
    }
    
    [dataToSend appendData:[packetQueue contentsForTransfer]];
    [packetQueue removeAllObjects];
    NSUInteger sendLength = [dataToSend length];
    NSUInteger written = [outputStream write:[dataToSend bytes] maxLength:sendLength];

    if(written == -1){
        [session peer:self encounteredReadWriteError:[outputStream streamError]];
    }
    
    if(written != sendLength){
        NSRange leftoverRange = NSMakeRange(written, [dataToSend length] - written);
        writeLeftover = [[dataToSend subdataWithRange:leftoverRange] retain];
    }
}

- (void)closeStreams
{
    if(inputStream){
        [inputStream close];
        [inputStream removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        [inputStream setDelegate:nil];
        [inputStream release];
        inputStream = nil;
    }
    
    if(outputStream){
        [outputStream close];
        [outputStream removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        [outputStream setDelegate:nil];
        [outputStream release];
        outputStream = nil;
    }
}

#pragma mark -
#pragma mark NSNetService Resolution

- (void)netServiceWillResolve:(NSNetService *)sender
{
    // nothing to do here - waiting for resolution
    // NSLog(@"resolving for %@.%@.%@", [sender name], [sender type], [sender domain]);
}

- (void)netService:(NSNetService *)sender didNotResolve:(NSDictionary *)errorDict
{
    [self setConnectionState:PSMLanPeerStateUnavailable];
    NSLog(@"failed resolving for %@:\n%@", [sender name], errorDict);
}

- (void)netServiceDidResolveAddress:(NSNetService *)sender
{
    NSLog(@"found remote service");
    // fetch the TXT record info to get the displayName for this service
    NSData *thisDisplayNameData = [[NSNetService dictionaryFromTXTRecordData:[sender TXTRecordData]] objectForKey:@"displayName"];
    NSString *thisDisplayName = [[NSString alloc] initWithData:thisDisplayNameData encoding:NSUTF8StringEncoding];
    [self setDisplayName:thisDisplayName];
    if(peerID && ![[session peerID] isEqualToString:peerID]) // don't report myself
        [self setConnectionState:PSMLanPeerStateAvailable];
    [thisDisplayName release];
}

#pragma mark -
#pragma mark Private Communication

- (BOOL)internalSendData:(NSData *)data
{
    if(data == nil || [data length] == 0)
        return NO;
    
    [packetQueue addObject:data];
    
    if([outputStream hasSpaceAvailable])
        [self sendQueuedData];
    
    return YES;
}

- (BOOL)sendMessage:(NSInteger)message withPayload:(NSData *)payload
{
    NSMutableData *sendData = [NSMutableData data];
    uint8_t signal[1];
    signal[0] = (uint8_t)message;
    [sendData appendBytes:signal length:sizeof(uint8_t)];
    if(payload)
        [sendData appendBytes:[payload bytes] length:[payload length]];
    return [self internalSendData:sendData];
}

- (void)handleMessage:(NSInteger)message withPayload:(NSData *)payload
{
    switch (message) {
            
        case kPSMPeerMessageAcceptRemoteConnection:
            // Yay - someone loves me
            [self killRequestTimer];
            if(connectionState == PSMLanPeerStateConnecting){
                [self setConnectionState:PSMLanPeerStateConnected];
            }
            break;
            
        case kPSMPeerMessageDenyRemoteConnection:
            // Ouch - I've been passed over
            [self killRequestTimer];
            if(connectionState == PSMLanPeerStateConnecting){
                // tell this end
                NSError *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionDeclinedError userInfo:nil];
                [session connectionRequestWithPeerFailed:self withError:error];
                
                [self closeStreams];
                [self revertState];
            }
            break;
            
        case kPSMPeerMessageDisconnect:
            [self closeStreams];
            [self setConnectionState:PSMLanPeerStateDisconnected];
            [self revertState];
            break;
            
        case kPSMPeerMessageACKRequest:
            [self sendACKResponse];
            break;
            
        case kPSMPeerMessageACKResponse:
            acknowledged = YES;
            break;
            
        case kPSMPeerMessageKickOutRemote:
            [session disconnectFromAllPeers];
            break;
            
        case kPSMPeerMessageSetErrorCode:
            NSLog(@"received error code");
            errorCode = (NSInteger)[payload bytes];
            break;
            
        default:
            break;
    }
}

- (void)acceptRemoteConnection
{
    if(![self sendMessage:kPSMPeerMessageAcceptRemoteConnection withPayload:nil])
        NSLog(@"Failed to send data to accept remote connection");
}

- (void)denyRemoteConnection
{
    if(![self sendMessage:kPSMPeerMessageDenyRemoteConnection withPayload:nil])
        NSLog(@"Failed to send data to deny remote connection");
}

- (void)disconnectRemote
{
    if(![self sendMessage:kPSMPeerMessageDisconnect withPayload:nil])
        NSLog(@"Failed to send data to disconnect remote connection");
}

- (void)requestRemoteACK
{
    if(![self sendMessage:kPSMPeerMessageACKRequest withPayload:nil])
        NSLog(@"Failed to send data to ACK Request remote connection");
}

- (void)sendACKResponse
{
    if(![self sendMessage:kPSMPeerMessageACKResponse withPayload:nil])
        NSLog(@"Failed to send data to ACK Response remote connection");
}

#pragma mark -
#pragma mark NSStream Delegate

- (void)stream:(NSStream *)aStream handleEvent:(NSStreamEvent)eventCode
{
    switch(eventCode){
        case NSStreamEventOpenCompleted:
            if(aStream == inputStream)
                readReady = YES;
            else
                writeReady = YES;
            break;
            
        case NSStreamEventHasBytesAvailable:
            if(aStream == inputStream){
                if([inputStream hasBytesAvailable]){
                    NSMutableData *data = [NSMutableData data];
                    
                    if(readLeftover != nil){
                        [data appendData:readLeftover];
                        [readLeftover release];
                        readLeftover = nil;
                    }
                    
                    NSInteger bytesRead;
                    uint8_t buffer[kBufferSize];
                    
                    bytesRead = [inputStream read:buffer maxLength:kBufferSize];
                    if(bytesRead == -1){
                        [session peer:self encounteredReadWriteError:[inputStream streamError]];
                        return;
                    } else if(bytesRead > 0){
                        [data appendBytes:buffer length:bytesRead];
                        
                        NSArray *dataPackets = [data splitTransferredPackets:&readLeftover];
                        if(readLeftover)
                            [readLeftover retain];
                        
                        for(NSData *item in dataPackets){
                            
                            // READ FIRST BYTE TO RESOLVE INTERNAL VS EXTERNAL COMMUNICATION
                            uint8_t signal[1];
                            [item getBytes:signal length:sizeof(uint8_t)];
                            NSRange messageRange = NSMakeRange(sizeof(uint8_t), [item length]-sizeof(uint8_t));
                            NSData *message = [item subdataWithRange:messageRange];
                            
                            if(signal[0] == kPSMPeerMessageExternal){
                                // data to pass through
                                [session peer:self receivedData:message];
                            } else {
                                // internal message - send to handler
                                [self handleMessage:(NSInteger)signal[0] withPayload:message];
                            }
                        }
                    }
                }
            }
            break;
            
        case NSStreamEventErrorOccurred:
            [self setConnectionState:PSMLanPeerStateDisconnected];
            [session peer:self encounteredReadWriteError:[aStream streamError]];
            break;
            
        case NSStreamEventHasSpaceAvailable:
            if(aStream == outputStream){
                [self sendQueuedData];
            }
            break;
            
        case NSStreamEventEndEncountered:
            writeReady = NO;
            readReady = NO;
            [self closeStreams];
            [self setConnectionState:PSMLanPeerStateAvailable];
            break;
            
        default:
            break;
    }
}

@end
