//
//  PSMLanSession.m
//  BombSquad
//
//  Created by John Pannell on 2/7/10.
//  Copyright 2010 Positive Spin Media. All rights reserved.
//

#import "PSMLanSession.h"
#import "NSString+UUID.h"
#import "PSMLanPeer.h"
#import "PSMLanListener.h"

@interface PSMLanSession()

// NSNetServiceBrowser delegate
- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didFindService:(NSNetService *)netService moreComing:(BOOL)moreServicesComing;
- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didNotSearch:(NSDictionary *)errorInfo;
- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didRemoveService:(NSNetService *)netService moreComing:(BOOL)moreServicesComing;
- (void)netServiceBrowserDidStopSearch:(NSNetServiceBrowser *)netServiceBrowser;
- (void)netServiceBrowserWillSearch:(NSNetServiceBrowser *)netServiceBrowser;

@end

@implementation PSMLanSession

@synthesize delegate, disconnectTimeout, displayName, peerID = _peerID, sessionID, sessionMode, availablePeers;
@dynamic available;

#pragma mark -
#pragma mark Creation/Destruction

- (id)initWithSessionID:(NSString *)newID displayName:(NSString *)name sessionMode:(PSMLanSessionMode)mode
{
    if(self = [super init]){
        sessionID = [newID retain];
        displayName = [name retain];
        sessionMode = mode;
        
        delegate = nil;
        peerLookup = [[NSMutableDictionary alloc] init];
        pendingPeers = [[NSMutableSet alloc] init];
        availablePeers = [[NSMutableSet alloc] init];
        _peerID = [[NSString stringWithNewUUID] retain];
        available = NO;
        disconnectTimeout = 20.0;
        broadcaster = nil;
        listener = nil;
        
        // looking for other peers
        serviceType = [[NSString alloc] initWithFormat:@"_%@._tcp.", sessionID];
        seeker = nil;
    }
    
    return self;
}

- (void)dealloc
{
    [sessionID release];
    [displayName release];
    [peerLookup release];
    [pendingPeers release];
    [availablePeers release];
    [_peerID release];
    [serviceType release];
    
    if(broadcaster){
        [broadcaster setDelegate:nil];
        [broadcaster stop];
        [broadcaster release];
    }
    
    if(listener){
        [listener release];
    }
    
    if(seeker){
        [seeker stop];
        [seeker release];
    }
    
    [super dealloc];
}

#pragma mark -
#pragma mark Bonjour

- (void)setAvailable:(BOOL)value
{
    if(value){ // I AM AVAILABLE!!
        if((sessionMode == PSMLanSessionModeServer) || (sessionMode == PSMLanSessionModePeer)){
            // Advertise presence and listen for connection requests
            listener = [[PSMLanListener alloc] init];
            [listener setDelegate:self];
            NSError *error;
            if(![listener startListening:&error]){
                NSLog(@"Listener error: %@", [error localizedDescription]);
                return;
            }
            
            broadcaster = [[NSNetService alloc] initWithDomain:@"local." type:serviceType name:[self peerID] port:[listener port]];
            
            // must broadcast my displayName
            NSDictionary *txtDict = [NSDictionary dictionaryWithObject:[[self displayName] dataUsingEncoding:NSUTF8StringEncoding] forKey:@"displayName"];
            NSData *txtData = [NSNetService dataFromTXTRecordDictionary:txtDict];
            [broadcaster setTXTRecordData:txtData];
            
            if(broadcaster){
                [broadcaster setDelegate:self];
                [broadcaster publish];
            } else {
                NSLog(@"An error occurred initializing the NSNetService object.");
            }
        }
        if((sessionMode == PSMLanSessionModeClient) || (sessionMode == PSMLanSessionModePeer)){
            // Seek out similar peers
            seeker = [[NSNetServiceBrowser alloc] init];
            [seeker setDelegate:self];
            [seeker searchForServicesOfType:serviceType inDomain:@"local."];
        }
    } else { // I AM NOT AVAILABLE
        //stop advertising presence via Bonjour and listening for connections
        if(broadcaster){
            [broadcaster stop];
            [broadcaster setDelegate:nil];
            [broadcaster release];
            broadcaster = nil;
        }
        if(listener){
            [listener stopListening];
            [listener release];
            listener = nil;
        }
        if(seeker){
            [seeker stop];
            [seeker release];
            seeker = nil;
        }
        // forget about any peers except those that are connected
        NSMutableArray *peersToRemove = [NSMutableArray array];
        for(NSString *key in peerLookup){
            if([[peerLookup objectForKey:key] connectionState] != PSMLanPeerStateConnected)
                [peersToRemove addObject:key];
        }
        [peerLookup removeObjectsForKeys:peersToRemove];
    }

    // record new value
    available = value;
}
        
#pragma mark -
#pragma mark Listener Delegate

- (void)acceptedConnectionForListener:(PSMLanListener *)theListener inputStream:(NSInputStream *)inStream outputStream:(NSOutputStream *)outStream address:(NSData *)address
{
    NSLog(@"%@", address);
    /* STRATEGY #1: USE IP ADDRESS TO FIND ALREADY PRESENT PEER AND HAND OVER STREAMS */
    /* NOTE: ONLY WORKS IN IPv4 */
    for(NSString *key in peerLookup){
        PSMLanPeer *thisPeer = [peerLookup objectForKey:key];
        NSArray *addresses = [[thisPeer remoteService] addresses];
        for(NSData *item in addresses){
            // address and item are the entire sockaddr struct - I just want to compare the address part (port will differ)
            NSData *itemIP = [item subdataWithRange:NSMakeRange(sizeof(uint32_t), sizeof(uint32_t))];
            NSData *addressIP = [address subdataWithRange:NSMakeRange(sizeof(uint32_t), sizeof(uint32_t))];

            if([itemIP isEqualToData:addressIP]){
                // found the right peer
                [inStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
                [outStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
                
                [inStream setDelegate:thisPeer];
                [outStream setDelegate:thisPeer];
                
                if([inStream streamStatus] == NSStreamStatusNotOpen)
                    [inStream open];
                if([outStream streamStatus] == NSStreamStatusNotOpen)
                    [outStream open];
                
                [thisPeer setInputStream:inStream];
                [thisPeer setOutputStream:outStream];
                
                // take action?
                [thisPeer setErrorCode:0]; // init this back, in case a previous connection attempt had failed
                [self peerDidReceiveConnectionRequest:thisPeer];
            }
        }
    }

    return;
}

#pragma mark -
#pragma mark Peer Delegate

- (void)peer:(PSMLanPeer *)peer didChangeState:(NSUInteger)newState
{
    NSString *thisPeerID = [[[peer peerID] copy] autorelease];
    
    // was this a pending peer? Move into collection if so...
    if(![peerLookup objectForKey:thisPeerID] && newState == PSMLanPeerStateAvailable){
        [peerLookup setObject:peer forKey:thisPeerID];
        [pendingPeers removeObject:peer];
    }
    
    if(newState == PSMLanPeerStateUnavailable){
        [availablePeers removeObject:thisPeerID];
        [peerLookup removeObjectForKey:thisPeerID];
    }
    
    if(delegate && [delegate respondsToSelector:@selector(session:peer:didChangeState:)]){
        [delegate session:self peer:thisPeerID didChangeState:newState];
    }
}

- (void)peerDidReceiveConnectionRequest:(PSMLanPeer *)peer
{    
    if(delegate && [delegate respondsToSelector:@selector(session:didReceiveConnectionRequestFromPeer:)]){
        [delegate session:self didReceiveConnectionRequestFromPeer:[peer peerID]];
    }
}

- (void)connectionRequestWithPeerFailed:(PSMLanPeer *)peer withError:(NSError *)error
{
    if(delegate && [delegate respondsToSelector:@selector(session:connectionWithPeerFailed:error:)]){
        [delegate session:self connectionWithPeerFailed:[peer peerID] withError:error];
    }
}
   
- (void)peer:(PSMLanPeer *)peer receivedData:(NSData *)data
{
    if(dataHandler && [dataHandler respondsToSelector:@selector(receiveData:fromPeer:inSession:context:)]){
        [dataHandler receiveData:data fromPeer:[peer peerID] inSession:self context:dataContext]; 
    }
}

- (void)peer:(PSMLanPeer *)peer encounteredReadWriteError:(NSError *)error
{
    // remove peer - no longer useful
    [peerLookup removeObjectForKey:[peer peerID]];
    
    if(delegate && [delegate respondsToSelector:@selector(session:connectionWithPeerFailed:)]){
        [delegate session:self connectionWithPeerFailed:[peer peerID] withError:error];
    }
}


#pragma mark -
#pragma mark Connecting to Peers

- (BOOL)acceptConnectionFromPeer:(NSString *)peerID error:(NSError **)error
{
    PSMLanPeer *thisPeer = [peerLookup objectForKey:peerID];
    if([thisPeer acceptConnection:error]){
        return YES;
    }
    
    return NO;
}

- (void)cancelConnectToPeer:(NSString *)peerID
{
    PSMLanPeer *thisPeer = [peerLookup objectForKey:peerID];
    [thisPeer cancelConnectionRequest];
}

- (void)connectToPeer:(NSString *)peerID withTimeout:(NSTimeInterval)timeout
{
    PSMLanPeer *thisPeer = [peerLookup objectForKey:peerID];
    [thisPeer initiateConnectionWithTimeout:timeout];
}

- (void)denyConnectionFromPeer:(NSString *)peerID
{
    PSMLanPeer *thisPeer = [peerLookup objectForKey:peerID];
    [thisPeer denyConnection];
}

- (void)disconnectFromAllPeers
{
    for(NSString *key in peerLookup){
        PSMLanPeer *aPeer = [peerLookup objectForKey:key];
        if([aPeer connectionState] == PSMLanPeerStateConnected){
            [aPeer performSelector:@selector(disconnect) withObject:nil afterDelay:0.0]; // concerned about mutation of collection caused by disconnect
        }
    }
}

- (void)disconnectPeerFromAllPeers:(NSString *)peerID
{
    PSMLanPeer *thisPeer = [peerLookup objectForKey:peerID];
    [thisPeer kickOut];
}

#pragma mark -
#pragma mark Information about peers

- (NSString *)displayNameForPeer:(NSString *)peerID
{
    return [[peerLookup objectForKey:peerID] displayName];
}

- (NSArray *)peersWithConnectionState:(PSMLanPeerConnectionState)state
{
    NSMutableArray *tempArray = [NSMutableArray array];
    for(NSString *key in peerLookup){
        PSMLanPeer *thisPeer = [peerLookup objectForKey:key];
        if([thisPeer connectionState] == state){
            [tempArray addObject:thisPeer];
        }
    }
    
    return [NSArray arrayWithArray:tempArray];
}

#pragma mark -
#pragma mark Data Exchange

- (BOOL)sendData:(NSData *)data toPeers:(NSArray *)peers withDataMode:(PSMLanSendDataMode)mode error:(NSError **)error
{
    BOOL success = YES;
    NSInteger sendCount = 0;
    for(NSString *aPeerID in peers){
        PSMLanPeer *aPeer = [peerLookup objectForKey:aPeerID];
        if(aPeer){
            if(!([aPeer connectionState] == PSMLanPeerStateConnected)){
                *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionNotConnectedError userInfo:nil];
                success = NO;
            } else {
                if(![aPeer sendData:data error:error] && success == YES){ // I'm ignoring the data mode - sends are reliable
                    success = NO;
                } else {
                    sendCount++;
                }
            }
        }
    }
    
    if(sendCount == 0)
        *error = [NSError errorWithDomain:PSMLanSessionErrorDomain code:PSMLanSessionPeerNotFoundError userInfo:nil];
    
    return success;
}

- (BOOL)sendDataToAllPeers:(NSData *)data withDataMode:(PSMLanSendDataMode)mode error:(NSError **)error
{
    NSMutableArray *peerGroup = [NSMutableArray array];
    for(NSString *key in peerLookup){
        PSMLanPeer *aPeer = [peerLookup objectForKey:key];
        if(aPeer && [aPeer connectionState] == PSMLanPeerStateConnected){
            [peerGroup addObject:[aPeer peerID]];
        }
    }
    
    return [self sendData:data toPeers:peerGroup withDataMode:mode error:error];
}

- (void)setDataReceiveHandler:(id)handler withContext:(void *)context
{
    dataHandler = handler;
    dataContext = context;
}

#pragma mark -
#pragma mark NSNetServiceBrowser delegate

- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didFindService:(NSNetService *)netService moreComing:(BOOL)moreServicesComing
{
    // Create a peer with this netservice
    PSMLanPeer *newPeer = [[PSMLanPeer alloc] initWithNetService:netService session:self];
    [pendingPeers addObject:newPeer];
    [newPeer release];
    
    // add to my collection of available peers
    [availablePeers addObject:[netService name]];
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didNotSearch:(NSDictionary *)errorInfo
{
    NSLog(@"Did not search: %@", errorInfo);
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)netServiceBrowser didRemoveService:(NSNetService *)netService moreComing:(BOOL)moreServicesComing
{
    NSLog(@"%@ no longer available", [netService name]);
    // no longer available
    [availablePeers removeObject:[netService name]];
    
    // I will also remove this from the lookup if it is not connected
    PSMLanPeer *relevantPeer = [peerLookup objectForKey:[netService name]];
    if(relevantPeer && [relevantPeer connectionState] != PSMLanPeerStateConnected){
        // ONLY FOR UNCONNECTED PEERS v v v 
        [relevantPeer hide];
        [peerLookup removeObjectForKey:[netService name]];
    }
}

- (void)netServiceBrowserDidStopSearch:(NSNetServiceBrowser *)netServiceBrowser
{
    NSLog(@"Searching stopped");
}

- (void)netServiceBrowserWillSearch:(NSNetServiceBrowser *)netServiceBrowser
{
    // NSLog(@"Searching for services");
}

@end
