//
//  NSInvocation+Serialization.m
//  BombSquad
//
//  Created by John Pannell on 2/9/10.
//  Copyright 2010 Positive Spin Media. All rights reserved.
//

#import "NSInvocation+Serialization.h"

@implementation NSInvocation(Serialization)

- (id)objectForArgumentAtIndex:(int)argIndex
{
    const char* argType;
    
    argType = [[self methodSignature] getArgumentTypeAtIndex:argIndex];
    while(strchr("rnNoORV", argType[0]) != NULL)
        argType += 1;
    
    if((strlen(argType) > 1) && (strchr("{^", argType[0]) == NULL))
        [NSException raise:NSInvalidArgumentException format:@"Cannot handle argument type '%s'.", argType];
    
    switch (argType[0]) 
    {
        case '@': 
        {
            id value;
            [self getArgument:&value atIndex:argIndex];
            return value;
        }
        case ':':
        {
            SEL s = (SEL)0;
            [self getArgument:&s atIndex:argIndex];
            id value = NSStringFromSelector(s);
            return value;
        }
        case 'i': 
        {
            int value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithInt:value];
        }	
        case 's':
        {
            short value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithShort:value];
        }	
        case 'l':
        {
            long value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithLong:value];
        }	
        case 'q':
        {
            long long value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithLongLong:value];
        }	
        case 'c':
        {
            char value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithChar:value];
        }	
        case 'C':
        {
            unsigned char value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithUnsignedChar:value];
        }	
        case 'I':
        {
            unsigned int value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithUnsignedInt:value];
        }	
        case 'S':
        {
            unsigned short value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithUnsignedShort:value];
        }	
        case 'L':
        {
            unsigned long value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithUnsignedLong:value];
        }	
        case 'Q':
        {
            unsigned long long value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithUnsignedLongLong:value];
        }	
        case 'f':
        {
            float value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithFloat:value];
        }	
        case 'd':
        {
            double value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithDouble:value];
        }	
        case 'B':
        {
            bool value;
            [self getArgument:&value atIndex:argIndex];
            return [NSNumber numberWithBool:value];
        }
    }
    [NSException raise:NSInvalidArgumentException format:@"Argument type '%s' not supported", argType];
    return nil;
}

- (NSData *)serializedData
{
    // I'm making an array of the arguments after index 0, and then making a binary plist data representation
    NSMutableArray *storage = [NSMutableArray array];
    NSUInteger numberOfArguments = [[self methodSignature] numberOfArguments];
    int i;
    
    // load array
    for(i = 1; i < numberOfArguments; i++){
        [storage addObject:[self objectForArgumentAtIndex:i]];
    }
    
    // serialize
    NSString *errorString;
    NSData *returnData = [NSPropertyListSerialization dataFromPropertyList:storage format:NSPropertyListBinaryFormat_v1_0 errorDescription:&errorString];
    if(errorString){
        [NSException raise:NSInvalidArgumentException format:@"Could not serialize invocation argument array: %@\nReason: %@", storage, errorString];
        [errorString release];
    } else {
        return returnData;
    }
    return nil;
}

@end

@implementation NSData(Serialization)

- (SEL)selectorForRepresentedInvocation
{
    // fish out the first element of the array and convert stored string to selector
    NSString *errorString;
    NSArray *storage = [NSPropertyListSerialization propertyListFromData:self mutabilityOption:0 format:NULL errorDescription:&errorString];
    if(errorString){
        [NSException raise:NSInvalidArgumentException format:@"Could not decode data for invocation: %@", errorString];
        [errorString release];
        return nil;
    }
    
    return NSSelectorFromString([storage objectAtIndex:0]);
}
    

- (NSInvocation *)representedInvocationForSignatature:(NSMethodSignature *)sig
{
    // decode a plist representation of the argument array and reconstruct
    NSString *errorString;
    NSArray *storage = [NSPropertyListSerialization propertyListFromData:self mutabilityOption:0 format:NULL errorDescription:&errorString];
    if(errorString){
        [NSException raise:NSInvalidArgumentException format:@"Could not decode data for invocation: %@", errorString];
        [errorString release];
        return nil;
    }
    
    // build up an NSInvocation
    NSInvocation *inv = [NSInvocation invocationWithMethodSignature:sig];
    [inv setSelector:NSSelectorFromString([storage objectAtIndex:0])];
    NSUInteger i, numArguments = [sig numberOfArguments];
    
    for(i = 2; i < numArguments; i++){
        id item = [storage objectAtIndex:(i-1)];
        
        // assign time
        const char* argType = [sig getArgumentTypeAtIndex:i];        
        switch (argType[0]) 
        {
            case '@': 
            {
                [inv setArgument:&item atIndex:i];
                break;
            }
            case 'i': 
            {
                int v = [item intValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 's':
            {
                short v = [item shortValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'l':
            {
                long v = [item longValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'q':
            {
                long long v = [item longLongValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'c':
            {
                char v = [item charValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'C':
            {
                unsigned char v = [item unsignedCharValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'I':
            {
                unsigned int v = [item unsignedIntValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'S':
            {
                unsigned short v = [item unsignedShortValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'L':
            {
                unsigned long v = [item unsignedLongValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'Q':
            {
                unsigned long long v = [item unsignedLongLongValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'f':
            {
                float v = [item floatValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'd':
            {
                double v = [item doubleValue];
                [inv setArgument:&v atIndex:i];
                break;
            }	
            case 'B':
            {
                bool v = [item boolValue];
                [inv setArgument:&v atIndex:i];
                break;
            }
        }
    }
    
    return inv;
}
    
    

@end
