//
//  PSMGameNode.m
//  BombSquad
//
//  Created by John Pannell on 2/10/10.
//  Copyright 2010 Positive Spin Media. All rights reserved.
//

#import "PSMGameNode.h"
#import "PSMLanSession.h"
#import "PSMPlayerProxy.h"
#import "NSInvocation+Serialization.h"
#if (TARGET_OS_IPHONE || TARGET_IPHONE || TARGET_IPHONE_SIMULATOR)
#import <GameKit/GameKit.h>
#endif

@interface PSMGameNode()

// session identification
- (BOOL)sessionIsLan:(id)session;

// deleagate methods for GK/LAN Sessions
- (void)session:(id)session connectionWithPeerFailed:(NSString *)peerID withError:(NSError *)error;
- (void)session:(id)session didFailWithError:(NSError *)error;
- (void)session:(id)session didReceiveConnectionRequestFromPeer:(NSString *)peerID;
- (void)session:(id)session peer:(NSString *)peerID didChangeState:(NSUInteger)state;

// getting data
- (void)receiveData:(NSData *)data fromPeer:(NSString *)peer inSession:(id)session context:(void *)context;

// errors sending messages with proxy (message is selector)
- (void)message:(NSString *)message toPlayer:(NSString *)playerID didFailWithError:(NSError *)error;

@end

@implementation PSMGameNode

@synthesize delegate, displayName;
@dynamic available, mode;

#pragma mark -
#pragma mark Creation/Destruction

- (id)initWithCommonIdentifier:(NSString *)identifier displayName:(NSString *)name mode:(PSMGameNodeMode)mode
{
    if(self = [super init]){
        
        displayName = [name retain];
        
        players = [[NSMutableDictionary alloc] init];
        sessions = [[NSMutableArray alloc] init];
        
        if(sessions){
            // add the LAN session first - preferred connection if both are available
            PSMLanSession *lanSession = [[PSMLanSession alloc] initWithSessionID:identifier displayName:name sessionMode:mode];
            [lanSession setDelegate:self];
            [lanSession setDataReceiveHandler:self withContext:nil];
            [sessions addObject:lanSession];
            [lanSession release];
            
            // if applicable, add a GKSession - SOMEDAY
            /*
            if(NSClassFromString(@"GKSession")){
                Class gkSessionClass = NSClassFromString(@"GKSession");
                id gkSession = [[gkSessionClass alloc] initWithSessionID:[NSString stringWithFormat:@"BT-%@", identifier] displayName:name sessionMode:mode];
                [gkSession setDelegate:self];
                [gkSession setDataReceiveHandler:self withContext:nil];
                [sessions addObject:gkSession];
                [gkSession release];
            }
             */
        }
    }
    
    return self;
}

- (void)dealloc
{
    [displayName release];
    [players release];
    [sessions release];
    
    [super dealloc];
}

#pragma mark -
#pragma mark Custom Accessors

- (BOOL)available
{
    return available;
}

- (void)setAvailable:(BOOL)value
{
    for(id item in sessions){
        [item setAvailable:value];
    }
    
    available = value;
}

- (BOOL)sessionIsLan:(id)session
{
    if([session isEqual:[sessions objectAtIndex:0]])
        return YES;
    
    return NO;
}

- (PSMGameNodeMode)mode
{
    return [[sessions objectAtIndex:0] sessionMode];
}


#pragma mark -
#pragma mark Session Delegate

- (void)session:(id)session connectionWithPeerFailed:(NSString *)peerID withError:(NSError *)error
{
    if(delegate && [delegate respondsToSelector:@selector(node:connectionWithPlayerFailed:withError:)]){
        [delegate node:self connectionWithPlayerFailed:peerID withError:error];
    }
}

- (void)session:(id)session didFailWithError:(NSError *)error
{
    if(delegate && [delegate respondsToSelector:@selector(node:didFailWithError:)]){
        [delegate node:self didFailWithError:error];
    }
}

- (void)session:(id)session didReceiveConnectionRequestFromPeer:(NSString *)peerID
{
    if(delegate && [delegate respondsToSelector:@selector(node:didReceiveConnectionRequestFromPlayer:)]){
        [delegate node:self didReceiveConnectionRequestFromPlayer:peerID];
    }
}

- (void)session:(id)session peer:(NSString *)peerID didChangeState:(NSUInteger)state
{
    BOOL found = NO;
    
    for(NSString *key in players){
        if([key isEqualToString:peerID]){
            found = YES; // it's already here - no need to create or swap out
            break;
        }
    }
                
    if(!found){
        NSInteger connType = PSMGameNodeConnectionTypeBluetooth;
        if([self sessionIsLan:session])
            connType = PSMGameNodeConnectionTypeLan;
        PSMPlayerProxy *newPlayer = [[PSMPlayerProxy alloc] initWithPlayerID:peerID displayName:[session displayNameForPeer:peerID] session:session connectionType:connType forNode:self];
        [players setObject:newPlayer forKey:peerID];
        [newPlayer release];
    }
    
    // relay message to delegate
    if(delegate && [delegate respondsToSelector:@selector(node:player:didChangeState:)]){
        [delegate node:self player:peerID didChangeState:state];
    }
}

- (void)receiveData:(NSData *)data fromPeer:(NSString *)peer inSession:(id)session context:(void *)context
{
    // the data received is a serialized NSInvocation.  Time to reconstitute it...
    SEL messageSelector = [data selectorForRepresentedInvocation];
    id target = nil;
    if(delegate && [delegate respondsToSelector:@selector(forwardingTargetForSelector:)]){
        target = [delegate forwardingTargetForSelector:messageSelector];
    }
    if(target){
        NSMethodSignature *signature = [target methodSignatureForSelector:messageSelector];
        NSInvocation *inv = [data representedInvocationForSignatature:signature];
        [inv setTarget:target];
        [inv invoke];
    }
}

- (void)message:(NSString *)message toPlayer:(NSString *)playerID didFailWithError:(NSError *)error
{
    if(delegate && [delegate respondsToSelector:@selector(message:toPlayer:didFailWithError:)]){
        [delegate message:message toPlayer:playerID didFailWithError:error];
    }
}

#pragma mark -
#pragma mark Connecting

- (void)connectToPlayer:(NSString *)playerID withTimeout:(NSTimeInterval)timeout
{
    id correctSession = [[self playerForPlayerID:playerID] hostSession];
    [correctSession connectToPeer:playerID withTimeout:timeout];
}

- (void)cancelConnectToPlayer:(NSString *)playerID
{
    id correctSession = [[self playerForPlayerID:playerID] hostSession];
    [correctSession cancelConnectToPeer:playerID];
}

- (BOOL)acceptConnectionFromPlayer:(NSString *)playerID error:(NSError **)error
{
    id correctSession = [[self playerForPlayerID:playerID] hostSession];
    return [correctSession acceptConnectionFromPeer:playerID error:error];
}

- (void)denyConnectionFromPlayer:(NSString *)playerID
{
    id correctSession = [[self playerForPlayerID:playerID] hostSession];
    [correctSession denyConnectionFromPeer:playerID];
}

- (void)disconnectFromAllPlayers
{
    for(id thisSession in sessions){
        [thisSession disconnectFromAllPeers];
    }
}

- (void)disconnectPlayerFromAllPlayers:(NSString *)playerID
{
    id correctSession = [[self playerForPlayerID:playerID] hostSession];
    [correctSession disconnectPeerFromAllPeers:playerID];
}

#pragma mark -
#pragma mark Usage

- (id)playerForPlayerID:(NSString *)playerID
{
    return [players objectForKey:playerID];
}

- (NSString *)displayNameForPlayerID:(NSString *)playerID
{
    return [[players objectForKey:playerID] displayName];
}

@end
